@@ -1,29 +1,24 @@
-"""
-Custom implementation of
+from collections import defaultdict
+from typing import Tuple, Sequence, List, Union, Dict
 
-    Shaw et al., 2019
-    MRI k-Space Motion Artefact Augmentation:
-    Model Robustness and Task-Specific Uncertainty
-
-"""
-
-import warnings
-from typing import Tuple, Optional, List
 import torch
 import numpy as np
 import SimpleITK as sitk
-from ....torchio import DATA, AFFINE
+
+from ....data.io import nib_to_sitk
 from ....data.subject import Subject
-from .. import Interpolation, get_sitk_interpolator
+from ....typing import TypeTripletFloat
+from ... import IntensityTransform, FourierTransform
 from .. import RandomTransform
 
 
-class RandomMotion(RandomTransform):
+class RandomMotion(RandomTransform, IntensityTransform, FourierTransform):
     r"""Add random MRI motion artifact.
 
-    Custom implementation of `Shaw et al. 2019, MRI k-Space Motion Artefact
-    Augmentation: Model Robustness and Task-Specific
-    Uncertainty <http://proceedings.mlr.press/v102/shaw19a.html>`_.
+    Magnetic resonance images suffer from motion artifacts when the subject
+    moves during image acquisition. This transform follows
+    `Shaw et al., 2019 <http://proceedings.mlr.press/v102/shaw19a.html>`_ to
+    simulate motion artifacts for data augmentation.
 
     Args:
         degrees: Tuple :math:`(a, b)` defining the rotation range in degrees of
@@ -43,86 +38,67 @@ class RandomMotion(RandomTransform):
         num_transforms: Number of simulated movements.
             Larger values generate more distorted images.
         image_interpolation: See :ref:`Interpolation`.
-        p: Probability that this transform will be applied.
-        seed: See :py:class:`~torchio.transforms.augmentation.RandomTransform`.
+        **kwargs: See :class:`~torchio.transforms.Transform` for additional
+            keyword arguments.
 
-    .. warning:: Large numbers of movements lead to longer execution times.
+    .. warning:: Large numbers of movements lead to longer execution times for
+        3D images.
     """
     def __init__(
             self,
             degrees: float = 10,
             translation: float = 10,  # in mm
             num_transforms: int = 2,
-            image_interpolation: Interpolation = Interpolation.LINEAR,
-            p: float = 1,
-            seed: Optional[int] = None,
+            image_interpolation: str = 'linear',
+            **kwargs
             ):
-        super().__init__(p=p, seed=seed)
+        super().__init__(**kwargs)
         self.degrees_range = self.parse_degrees(degrees)
         self.translation_range = self.parse_translation(translation)
+        if not 0 < num_transforms or not isinstance(num_transforms, int):
+            message = (
+                'Number of transforms must be a strictly positive natural'
+                f'number, not {num_transforms}'
+            )
+            raise ValueError(message)
         self.num_transforms = num_transforms
-        self.image_interpolation = image_interpolation
+        self.image_interpolation = self.parse_interpolation(
+            image_interpolation)
 
-    def apply_transform(self, sample: Subject) -> dict:
-        random_parameters_images_dict = {}
-        for image_name, image_dict in sample.get_images_dict().items():
+    def apply_transform(self, subject: Subject) -> Subject:
+        arguments = defaultdict(dict)
+        for name, image in self.get_images_dict(subject).items():
             params = self.get_params(
                 self.degrees_range,
                 self.translation_range,
                 self.num_transforms,
+                is_2d=image.is_2d(),
             )
             times_params, degrees_params, translation_params = params
-            random_parameters_dict = {
-                'times': times_params,
-                'degrees': degrees_params,
-                'translation': translation_params,
-            }
-            random_parameters_images_dict[image_name] = random_parameters_dict
-            if (image_dict[DATA][0] < -0.1).any():
-                # I use -0.1 instead of 0 because Python was warning me when
-                # a value in a voxel was -7.191084e-35
-                # There must be a better way of solving this
-                message = (
-                    f'Image "{image_name}" from "{image_dict["stem"]}"'
-                    ' has negative values.'
-                    ' Results can be unexpected because the transformed sample'
-                    ' is computed as the absolute values'
-                    ' of an inverse Fourier transform'
-                )
-                warnings.warn(message)
-            image = self.nib_to_sitk(
-                image_dict[DATA][0],
-                image_dict[AFFINE],
-            )
-            transforms = self.get_rigid_transforms(
-                degrees_params,
-                translation_params,
-                image,
-            )
-            image_dict[DATA] = self.add_artifact(
-                image,
-                transforms,
-                times_params,
-                self.image_interpolation,
-            )
-            # Add channels dimension
-            image_dict[DATA] = image_dict[DATA][np.newaxis, ...]
-            image_dict[DATA] = torch.from_numpy(image_dict[DATA])
-        sample.add_transform(self, random_parameters_images_dict)
-        return sample
+            arguments['times'][name] = times_params
+            arguments['degrees'][name] = degrees_params
+            arguments['translation'][name] = translation_params
+            arguments['image_interpolation'][name] = self.image_interpolation
+        transform = Motion(**self.add_include_exclude(arguments))
+        transformed = transform(subject)
+        return transformed
 
-    @staticmethod
     def get_params(
+            self,
             degrees_range: Tuple[float, float],
             translation_range: Tuple[float, float],
             num_transforms: int,
             perturbation: float = 0.3,
-            ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, bool]:
+            is_2d: bool = False,
+            ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
         # If perturbation is 0, time intervals between movements are constant
-        degrees_params = get_params_array(
+        degrees_params = self.get_params_array(
             degrees_range, num_transforms)
-        translation_params = get_params_array(
+        translation_params = self.get_params_array(
             translation_range, num_transforms)
+        if is_2d:  # imagine sagittal (1, A, S)
+            degrees_params[:, :-1] = 0  # rotate around Z axis only
+            translation_params[:, 2] = 0  # translate in XY plane only
         step = 1 / (num_transforms + 1)
         times = torch.arange(0, 1, step)[1:]
         noise = torch.FloatTensor(num_transforms)
@@ -131,6 +107,82 @@ class RandomMotion(RandomTransform):
         times_params = times.numpy()
         return times_params, degrees_params, translation_params
 
+    @staticmethod
+    def get_params_array(nums_range: Tuple[float, float], num_transforms: int):
+        tensor = torch.FloatTensor(num_transforms, 3).uniform_(*nums_range)
+        return tensor.numpy()
+
+
+class Motion(IntensityTransform, FourierTransform):
+    r"""Add MRI motion artifact.
+
+    Magnetic resonance images suffer from motion artifacts when the subject
+    moves during image acquisition. This transform follows
+    `Shaw et al., 2019 <http://proceedings.mlr.press/v102/shaw19a.html>`_ to
+    simulate motion artifacts for data augmentation.
+
+    Args:
+        degrees: Sequence of rotations :math:`(\theta_1, \theta_2, \theta_3)`.
+        translation: Sequence of translations :math:`(t_1, t_2, t_3)` in mm.
+        times: Sequence of times from 0 to 1 at which the motions happen.
+        image_interpolation: See :ref:`Interpolation`.
+        **kwargs: See :class:`~torchio.transforms.Transform` for additional
+            keyword arguments.
+    """
+    def __init__(
+            self,
+            degrees: Union[TypeTripletFloat, Dict[str, TypeTripletFloat]],
+            translation: Union[TypeTripletFloat, Dict[str, TypeTripletFloat]],
+            times: Union[Sequence[float], Dict[str, Sequence[float]]],
+            image_interpolation: Union[Sequence[str], Dict[str, Sequence[str]]],  # noqa: E501
+            **kwargs
+            ):
+        super().__init__(**kwargs)
+        self.degrees = degrees
+        self.translation = translation
+        self.times = times
+        self.image_interpolation = image_interpolation
+        self.args_names = (
+            'degrees',
+            'translation',
+            'times',
+            'image_interpolation',
+        )
+
+    def apply_transform(self, subject: Subject) -> Subject:
+        degrees = self.degrees
+        translation = self.translation
+        times = self.times
+        image_interpolation = self.image_interpolation
+        for image_name, image in self.get_images_dict(subject).items():
+            if self.arguments_are_dict():
+                degrees = self.degrees[image_name]
+                translation = self.translation[image_name]
+                times = self.times[image_name]
+                image_interpolation = self.image_interpolation[image_name]
+            result_arrays = []
+            for channel in image.data:
+                sitk_image = nib_to_sitk(
+                    channel[np.newaxis],
+                    image.affine,
+                    force_3d=True,
+                )
+                transforms = self.get_rigid_transforms(
+                    degrees,
+                    translation,
+                    sitk_image,
+                )
+                transformed_channel = self.add_artifact(
+                    sitk_image,
+                    transforms,
+                    times,
+                    image_interpolation,
+                )
+                result_arrays.append(transformed_channel)
+            result = np.stack(result_arrays)
+            image.set_data(torch.from_numpy(result))
+        return subject
+
     def get_rigid_transforms(
             self,
             degrees_params: np.ndarray,
@@ -168,19 +220,20 @@ class RandomMotion(RandomTransform):
         transform.SetTranslation(matrix[:3, 3])
         return transform
 
-    @staticmethod
     def resample_images(
+            self,
             image: sitk.Image,
-            transforms: List[sitk.Euler3DTransform],
-            interpolation: Interpolation,
+            transforms: Sequence[sitk.Euler3DTransform],
+            interpolation: str,
             ) -> List[sitk.Image]:
         floating = reference = image
         default_value = np.float64(sitk.GetArrayViewFromImage(image).min())
         transforms = transforms[1:]  # first is identity
         images = [image]  # first is identity
         for transform in transforms:
+            interpolator = self.get_sitk_interpolator(interpolation)
             resampler = sitk.ResampleImageFilter()
-            resampler.SetInterpolator(get_sitk_interpolator(interpolation))
+            resampler.SetInterpolator(interpolator)
             resampler.SetReferenceImage(reference)
             resampler.SetOutputPixelType(sitk.sitkFloat32)
             resampler.SetDefaultPixelValue(default_value)
@@ -202,9 +255,9 @@ class RandomMotion(RandomTransform):
     def add_artifact(
             self,
             image: sitk.Image,
-            transforms: List[sitk.Euler3DTransform],
+            transforms: Sequence[sitk.Euler3DTransform],
             times: np.ndarray,
-            interpolation: Interpolation,
+            interpolation: str,
             ):
         images = self.resample_images(image, transforms, interpolation)
         arrays = [sitk.GetArrayViewFromImage(im) for im in images]
@@ -219,10 +272,5 @@ class RandomMotion(RandomTransform):
         for spectrum, fin in zip(spectra, indices):
             result_spectrum[..., ini:fin] = spectrum[..., ini:fin]
             ini = fin
-        result_image = self.inv_fourier_transform(result_spectrum)
+        result_image = np.real(self.inv_fourier_transform(result_spectrum))
         return result_image.astype(np.float32)
-
-
-def get_params_array(nums_range: Tuple[float, float], num_transforms: int):
-    tensor = torch.FloatTensor(num_transforms, 3).uniform_(*nums_range)
-    return tensor.numpy()
