@@ -1,20 +1,24 @@
 import warnings
 from numbers import Number
-from typing import Tuple, Optional, Union
+from typing import Tuple, Union, Sequence
+
 import torch
 import numpy as np
 import SimpleITK as sitk
-from ....data.subject import Subject
+
 from ....utils import to_tuple
-from ....torchio import LABEL, DATA, AFFINE, TYPE
-from .. import Interpolation, get_sitk_interpolator
+from ....data.io import nib_to_sitk
+from ....data.subject import Subject
+from ....data.image import ScalarImage
+from ....typing import TypeTripletInt, TypeTripletFloat
+from ... import SpatialTransform
 from .. import RandomTransform
 
 
 SPLINE_ORDER = 3
 
 
-class RandomElasticDeformation(RandomTransform):
+class RandomElasticDeformation(RandomTransform, SpatialTransform):
     r"""Apply dense random elastic deformation.
 
     A random displacement is assigned to a coarse grid of control points around
@@ -43,19 +47,17 @@ class RandomElasticDeformation(RandomTransform):
             :math:`D_{max} = \sqrt{D_x^2 + D_y^2 + D_z^2}`.
         locked_borders: If ``0``, all displacement vectors are kept.
             If ``1``, displacement of control points at the
-            border of the coarse grid will also be set to ``0``.
+            border of the coarse grid will be set to ``0``.
             If ``2``, displacement of control points at the border of the image
-            will also be set to ``0``.
-        image_interpolation: Value in
-            :py:class:`torchio.transforms.interpolation.Interpolation` (see
-            :ref:`Interpolation`).
+            (red dots in the image below) will also be set to ``0``.
+        image_interpolation: See :ref:`Interpolation`.
             Note that this is the interpolation used to compute voxel
             intensities when resampling using the dense displacement field.
             The value of the dense displacement at each voxel is always
             interpolated with cubic B-splines from the values at the control
             points of the coarse grid.
-        p: Probability that this transform will be applied.
-        seed: See :py:class:`~torchio.transforms.augmentation.RandomTransform`.
+        **kwargs: See :class:`~torchio.transforms.Transform` for additional
+            keyword arguments.
 
     `This gist <https://gist.github.com/fepegar/b723d15de620cd2a3a4dbd71e491b59d>`_
     can also be used to better understand the meaning of the parameters.
@@ -76,7 +78,7 @@ class RandomElasticDeformation(RandomTransform):
         ... )
 
     Note that control points outside the image bounds are not showed in the
-    example image (they would also be red as we set :py:attr:`locked_borders`
+    example image (they would also be red as we set :attr:`locked_borders`
     to ``2``).
 
     .. warning:: Image folding may occur if the maximum displacement is larger
@@ -85,45 +87,45 @@ class RandomElasticDeformation(RandomTransform):
         points::
 
             >>> import numpy as np
-            >>> import SimpleITK as sitk
-            >>> image = sitk.ReadImage('my_image.nii.gz')
-            >>> image.GetSize()
-            (512, 512, 139)  # voxels
-            >>> image.GetSpacing()
-            (0.76, 0.76, 2.50)  # mm
+            >>> import torchio as tio
+            >>> image = tio.datasets.Slicer().MRHead.as_sitk()
+            >>> image.GetSize()  # in voxels
+            (256, 256, 130)
+            >>> image.GetSpacing()  # in mm
+            (1.0, 1.0, 1.2999954223632812)
             >>> bounds = np.array(image.GetSize()) * np.array(image.GetSpacing())
-            array([390.0, 390.0, 347.5])  # mm
+            >>> bounds  # mm
+            array([256.        , 256.        , 168.99940491])
             >>> num_control_points = np.array((7, 7, 6))
             >>> grid_spacing = bounds / (num_control_points - 2)
             >>> grid_spacing
-            array([78.0, 78.0, 86.9])  # mm
+            array([51.2       , 51.2       , 42.24985123])
             >>> potential_folding = grid_spacing / 2
-            >>> potential_folding
-            array([39.0, 39.0, 43.4])  # mm
+            >>> potential_folding  # mm
+            array([25.6       , 25.6       , 21.12492561])
 
-        Using a :py:attr:`max_displacement` larger than the computed
-        :py:attr:`potential_folding` will raise a :py:class:`RuntimeWarning`.
+        Using a :attr:`max_displacement` larger than the computed
+        :attr:`potential_folding` will raise a :class:`RuntimeWarning`.
 
         .. [#] Technically, :math:`2 \epsilon` should be added to the
             image bounds, where :math:`\epsilon = 2^{-3}` `according to ITK
             source code <https://github.com/InsightSoftwareConsortium/ITK/blob/633f84548311600845d54ab2463d3412194690a8/Modules/Core/Transform/include/itkBSplineTransformInitializer.hxx#L116-L138>`_.
-    """
+    """  # noqa: E501
 
     def __init__(
             self,
             num_control_points: Union[int, Tuple[int, int, int]] = 7,
             max_displacement: Union[float, Tuple[float, float, float]] = 7.5,
             locked_borders: int = 2,
-            image_interpolation: Interpolation = Interpolation.LINEAR,
-            p: float = 1,
-            seed: Optional[int] = None,
+            image_interpolation: str = 'linear',
+            **kwargs
             ):
-        super().__init__(p=p, seed=seed)
+        super().__init__(**kwargs)
         self._bspline_transformation = None
         self.num_control_points = to_tuple(num_control_points, length=3)
-        self.parse_control_points(self.num_control_points)
+        _parse_num_control_points(self.num_control_points)
         self.max_displacement = to_tuple(max_displacement, length=3)
-        self.parse_max_displacement(self.max_displacement)
+        _parse_max_displacement(self.max_displacement)
         self.num_locked_borders = locked_borders
         if locked_borders not in (0, 1, 2):
             raise ValueError('locked_borders must be 0, 1, or 2')
@@ -134,39 +136,15 @@ class RandomElasticDeformation(RandomTransform):
                 ' or use more control points.'
             )
             raise ValueError(message)
-        self.interpolation = self.parse_interpolation(image_interpolation)
-
-    @staticmethod
-    def parse_control_points(
-            num_control_points: Tuple[int, int, int],
-            ) -> None:
-        for axis, number in enumerate(num_control_points):
-            if not isinstance(number, int) or number < 4:
-                message = (
-                    f'The number of control points for axis {axis} must be'
-                    f' an integer greater than 3, not {number}'
-                )
-                raise ValueError(message)
-
-    @staticmethod
-    def parse_max_displacement(
-            max_displacement: Tuple[float, float, float],
-            ) -> None:
-        for axis, number in enumerate(max_displacement):
-            if not isinstance(number, Number) or number < 0:
-                message = (
-                    'The maximum displacement at each control point'
-                    f' for axis {axis} must be'
-                    f' a number greater or equal to 0, not {number}'
-                )
-                raise ValueError(message)
+        self.image_interpolation = self.parse_interpolation(
+            image_interpolation)
 
     @staticmethod
     def get_params(
-            num_control_points: Tuple[int, int, int],
+            num_control_points: TypeTripletInt,
             max_displacement: Tuple[float, float, float],
             num_locked_borders: int,
-            ) -> Tuple:
+            ) -> np.ndarray:
         grid_shape = num_control_points
         num_dimensions = 3
         coarse_field = torch.rand(*grid_shape, num_dimensions)  # [0, 1)
@@ -185,20 +163,72 @@ class RandomElasticDeformation(RandomTransform):
 
         return coarse_field.numpy()
 
+    def apply_transform(self, subject: Subject) -> Subject:
+        subject.check_consistent_spatial_shape()
+        control_points = self.get_params(
+            self.num_control_points,
+            self.max_displacement,
+            self.num_locked_borders,
+        )
+
+        arguments = {
+            'control_points': control_points,
+            'max_displacement': self.max_displacement,
+            'image_interpolation': self.image_interpolation,
+        }
+
+        transform = ElasticDeformation(**self.add_include_exclude(arguments))
+        transformed = transform(subject)
+        return transformed
+
+
+class ElasticDeformation(SpatialTransform):
+    r"""Apply dense elastic deformation.
+
+    Args:
+        control_points:
+        max_displacement:
+        image_interpolation: See :ref:`Interpolation`.
+        **kwargs: See :class:`~torchio.transforms.Transform` for additional
+            keyword arguments.
+    """
+
+    def __init__(
+            self,
+            control_points: np.ndarray,
+            max_displacement: TypeTripletFloat,
+            image_interpolation: str = 'linear',
+            **kwargs
+            ):
+        super().__init__(**kwargs)
+        self.control_points = control_points
+        self.max_displacement = max_displacement
+        self.image_interpolation = self.parse_interpolation(
+            image_interpolation)
+        self.invert_transform = False
+        self.args_names = (
+            'control_points',
+            'image_interpolation',
+            'max_displacement',
+        )
+
     @staticmethod
     def get_bspline_transform(
             image: sitk.Image,
-            num_control_points: Tuple[int, int, int],
-            coarse_field: np.ndarray,
+            control_points: np.ndarray,
             ) -> sitk.BSplineTransformInitializer:
+        num_control_points = control_points.shape[:-1]
         mesh_shape = [n - SPLINE_ORDER for n in num_control_points]
         bspline_transform = sitk.BSplineTransformInitializer(image, mesh_shape)
-        parameters = coarse_field.flatten(order='F').tolist()
+        parameters = control_points.flatten(order='F').tolist()
         bspline_transform.SetParameters(parameters)
         return bspline_transform
 
     @staticmethod
-    def parse_free_form_transform(transform, max_displacement):
+    def parse_free_form_transform(
+            transform: sitk.Transform,
+            max_displacement: Sequence[TypeTripletInt],
+            ) -> None:
         """Issue a warning is possible folding is detected."""
         coefficient_images = transform.GetCoefficientImages()
         grid_spacing = coefficient_images[0].GetSpacing()
@@ -213,55 +243,84 @@ class RandomElasticDeformation(RandomTransform):
             )
             warnings.warn(message, RuntimeWarning)
 
-    def apply_transform(self, sample: Subject) -> dict:
-        sample.check_consistent_shape()
-        bspline_params = self.get_params(
-            self.num_control_points,
-            self.max_displacement,
-            self.num_locked_borders,
-        )
-        random_parameters_dict = {'coarse_grid': bspline_params}
-        for image_dict in sample.get_images(intensity_only=False):
-            if image_dict[TYPE] == LABEL:
-                interpolation = Interpolation.NEAREST
+    def apply_transform(self, subject: Subject) -> Subject:
+        no_displacement = not any(self.max_displacement)
+        if no_displacement:
+            return subject
+        subject.check_consistent_spatial_shape()
+        control_points = self.control_points.copy()
+        if self.invert_transform:
+            control_points *= -1
+        for image in self.get_images(subject):
+            if not isinstance(image, ScalarImage):
+                interpolation = 'nearest'
             else:
-                interpolation = self.interpolation
-            image_dict[DATA] = self.apply_bspline_transform(
-                image_dict[DATA],
-                image_dict[AFFINE],
-                bspline_params,
+                interpolation = self.image_interpolation
+            if image.is_2d():
+                control_points[..., -1] = 0  # no displacement in IS axis
+            transformed = self.apply_bspline_transform(
+                image.data,
+                image.affine,
+                control_points,
                 interpolation,
             )
-        sample.add_transform(self, random_parameters_dict)
-        return sample
+            image.set_data(transformed)
+        return subject
 
     def apply_bspline_transform(
             self,
             tensor: torch.Tensor,
             affine: np.ndarray,
-            bspline_params: np.ndarray,
-            interpolation: Interpolation,
+            control_points: np.ndarray,
+            interpolation: str,
             ) -> torch.Tensor:
-        assert tensor.ndim == 4
-        assert len(tensor) == 1
-        image = self.nib_to_sitk(tensor[0], affine)
-        floating = reference = image
-        bspline_transform = self.get_bspline_transform(
-            image,
-            self.num_control_points,
-            bspline_params,
-        )
-        self.parse_free_form_transform(
-            bspline_transform, self.max_displacement)
-        resampler = sitk.ResampleImageFilter()
-        resampler.SetReferenceImage(reference)
-        resampler.SetTransform(bspline_transform)
-        resampler.SetInterpolator(get_sitk_interpolator(interpolation))
-        resampler.SetDefaultPixelValue(tensor.min().item())
-        resampler.SetOutputPixelType(sitk.sitkFloat32)
-        resampled = resampler.Execute(floating)
-
-        np_array = sitk.GetArrayFromImage(resampled)
-        np_array = np_array.transpose()  # ITK to NumPy
-        tensor[0] = torch.from_numpy(np_array)
+        assert tensor.dim() == 4
+        results = []
+        for component in tensor:
+            image = nib_to_sitk(component[np.newaxis], affine, force_3d=True)
+            floating = reference = image
+            bspline_transform = self.get_bspline_transform(
+                image,
+                control_points,
+            )
+            self.parse_free_form_transform(
+                bspline_transform,
+                self.max_displacement,
+            )
+            interpolator = self.get_sitk_interpolator(interpolation)
+            resampler = sitk.ResampleImageFilter()
+            resampler.SetReferenceImage(reference)
+            resampler.SetTransform(bspline_transform)
+            resampler.SetInterpolator(interpolator)
+            resampler.SetDefaultPixelValue(component.min().item())
+            resampler.SetOutputPixelType(sitk.sitkFloat32)
+            resampled = resampler.Execute(floating)
+            result, _ = self.sitk_to_nib(resampled)
+            results.append(torch.from_numpy(result))
+        tensor = torch.cat(results)
         return tensor
+
+
+def _parse_num_control_points(
+        num_control_points: TypeTripletInt,
+        ) -> None:
+    for axis, number in enumerate(num_control_points):
+        if not isinstance(number, int) or number < 4:
+            message = (
+                f'The number of control points for axis {axis} must be'
+                f' an integer greater than 3, not {number}'
+            )
+            raise ValueError(message)
+
+
+def _parse_max_displacement(
+        max_displacement: Tuple[float, float, float],
+        ) -> None:
+    for axis, number in enumerate(max_displacement):
+        if not isinstance(number, Number) or number < 0:
+            message = (
+                'The maximum displacement at each control point'
+                f' for axis {axis} must be'
+                f' a number greater or equal to 0, not {number}'
+            )
+            raise ValueError(message)
