@@ -1,23 +1,18 @@
 import ast
+import gzip
 import shutil
 import tempfile
 from pathlib import Path
-from typing import Union, Iterable, Tuple, Any, Optional, List
-import torch
+from typing import Union, Iterable, Tuple, Any, Optional, List, Sequence
+
+from torch.utils.data._utils.collate import default_collate
 import numpy as np
 import nibabel as nib
 import SimpleITK as sitk
 from tqdm import trange
-from .torchio import (
-    INTENSITY,
-    LABEL,
-    TypeData,
-    TypeNumber,
-    TypePath,
-)
 
-
-FLIP_XY = np.diag((-1, -1, 1))
+from .constants import INTENSITY
+from .typing import TypeNumber, TypePath
 
 
 def to_tuple(
@@ -41,12 +36,17 @@ def to_tuple(
     return value
 
 
-def get_stem(path: TypePath) -> str:
+def get_stem(
+        path: Union[TypePath, List[TypePath]]
+        ) -> Union[str, List[str]]:
     """
     '/home/user/image.nii.gz' -> 'image'
     """
-    path = Path(path)
-    return path.name.split('.')[0]
+    def _get_stem(path_string):
+        return Path(path_string).name.split('.')[0]
+    if isinstance(path, (str, Path)):
+        return _get_stem(path)
+    return [_get_stem(p) for p in path]
 
 
 def create_dummy_dataset(
@@ -57,7 +57,7 @@ def create_dummy_dataset(
         force: bool = False,
         verbose: bool = False,
         ):
-    from .data import Image, Subject
+    from .data import ScalarImage, LabelMap, Subject
     output_dir = tempfile.gettempdir() if directory is None else directory
     output_dir = Path(output_dir)
     images_dir = output_dir / 'dummy_images'
@@ -73,15 +73,15 @@ def create_dummy_dataset(
             image_path = images_dir / f'image_{i}{suffix}'
             label_path = labels_dir / f'label_{i}{suffix}'
             subject = Subject(
-                one_modality=Image(image_path, INTENSITY),
-                segmentation=Image(label_path, LABEL),
+                one_modality=ScalarImage(image_path),
+                segmentation=LabelMap(label_path),
             )
             subjects.append(subject)
     else:
         images_dir.mkdir(exist_ok=True, parents=True)
         labels_dir.mkdir(exist_ok=True, parents=True)
         if verbose:
-            print('Creating dummy dataset...')
+            print('Creating dummy dataset...')  # noqa: T001
             iterable = trange(num_images)
         else:
             iterable = range(num_images)
@@ -103,8 +103,8 @@ def create_dummy_dataset(
             nii.to_filename(str(label_path))
 
             subject = Subject(
-                one_modality=Image(image_path, INTENSITY),
-                segmentation=Image(label_path, LABEL),
+                one_modality=ScalarImage(image_path),
+                segmentation=LabelMap(label_path),
             )
             subjects.append(subject)
     return subjects
@@ -112,22 +112,22 @@ def create_dummy_dataset(
 
 def apply_transform_to_file(
         input_path: TypePath,
-        transform,  # : Transform seems to create a circular import (TODO)
+        transform,  # : Transform seems to create a circular import
         output_path: TypePath,
-        type_: str = INTENSITY,
+        type: str = INTENSITY,  # noqa: A002
+        verbose: bool = False,
         ):
-    from . import Image, ImagesDataset, Subject
-    subject = Subject(image=Image(input_path, type_))
-    dataset = ImagesDataset([subject], transform=transform)
-    transformed = dataset[0]
-    dataset.save_sample(transformed, dict(image=output_path))
+    from . import Image, Subject
+    subject = Subject(image=Image(input_path, type=type))
+    transformed = transform(subject)
+    transformed.image.save(output_path)
+    if verbose and transformed.history:
+        print('Applied transform:', transformed.history[0])  # noqa: T001
 
 
 def guess_type(string: str) -> Any:
-    """
-    Adapted from
-    https://www.reddit.com/r/learnpython/comments/4599hl/module_to_guess_type_from_a_string/czw3f5s
-    """
+    # Adapted from
+    # https://www.reddit.com/r/learnpython/comments/4599hl/module_to_guess_type_from_a_string/czw3f5s
     string = string.replace(' ', '')
     try:
         value = ast.literal_eval(string)
@@ -148,42 +148,11 @@ def guess_type(string: str) -> Any:
     return value
 
 
-def get_rotation_and_spacing_from_affine(
-        affine: np.ndarray,
-        ) -> Tuple[np.ndarray, np.ndarray]:
-    rotation_zoom = affine[:3, :3]
-    spacing = np.sqrt(np.sum(rotation_zoom * rotation_zoom, axis=0))
-    rotation = rotation_zoom / spacing
-    return rotation, spacing
+def get_torchio_cache_dir():
+    return Path('~/.cache/torchio').expanduser()
 
 
-def nib_to_sitk(data: TypeData, affine: TypeData) -> sitk.Image:
-    array = data.numpy() if isinstance(data, torch.Tensor) else data
-    affine = affine.numpy() if isinstance(affine, torch.Tensor) else affine
-    origin = np.dot(FLIP_XY, affine[:3, 3]).astype(np.float64)
-    rotation, spacing = get_rotation_and_spacing_from_affine(affine)
-    direction = np.dot(FLIP_XY, rotation).flatten()
-    image = sitk.GetImageFromArray(array.transpose())
-    image.SetOrigin(origin)
-    image.SetSpacing(spacing)
-    image.SetDirection(direction)
-    return image
-
-
-def sitk_to_nib(image: sitk.Image) -> Tuple[np.ndarray, np.ndarray]:
-    data = sitk.GetArrayFromImage(image).transpose()
-    spacing = np.array(image.GetSpacing())
-    rotation = np.array(image.GetDirection()).reshape(3, 3)
-    rotation = np.dot(FLIP_XY, rotation)
-    rotation_zoom = rotation * spacing
-    translation = np.dot(FLIP_XY, image.GetOrigin())
-    affine = np.eye(4)
-    affine[:3, :3] = rotation_zoom
-    affine[:3, 3] = translation
-    return data, affine
-
-
-def round_up(value: float) -> float:
+def round_up(value: float) -> int:
     """Round half towards infinity.
 
     Args:
@@ -201,4 +170,48 @@ def round_up(value: float) -> float:
         4
 
     """
-    return np.floor(value + 0.5)
+    return int(np.floor(value + 0.5))
+
+
+def compress(input_path, output_path):
+    with open(input_path, 'rb') as f_in:
+        with gzip.open(output_path, 'wb') as f_out:
+            shutil.copyfileobj(f_in, f_out)
+
+
+def check_sequence(sequence: Sequence, name: str):
+    try:
+        iter(sequence)
+    except TypeError:
+        message = f'"{name}" must be a sequence, not {type(name)}'
+        raise TypeError(message)
+
+
+def get_major_sitk_version() -> int:
+    # This attribute was added in version 2
+    # https://github.com/SimpleITK/SimpleITK/pull/1171
+    version = getattr(sitk, '__version__', None)
+    major_version = 1 if version is None else 2
+    return major_version
+
+
+def history_collate(batch: Sequence, collate_transforms=True):
+    attr = 'history' if collate_transforms else 'applied_transforms'
+    # Adapted from
+    # https://github.com/romainVala/torchQC/blob/master/segmentation/collate_functions.py
+    from .data import Subject
+    first_element = batch[0]
+    if isinstance(first_element, Subject):
+        dictionary = {
+            key: default_collate([d[key] for d in batch])
+            for key in first_element
+        }
+        if hasattr(first_element, attr):
+            dictionary.update({attr: [getattr(d, attr) for d in batch]})
+        return dictionary
+
+
+def get_subclasses(target_class: type) -> List[type]:
+    subclasses = target_class.__subclasses__()
+    subclasses += sum((get_subclasses(cls) for cls in subclasses), [])
+    return subclasses
